#include "config.h"
#include "bme280.h"
#include "main.h"

#include <fsl_i2c_master_driver.h>

static i2c_master_state_t i2c;
static BMP_calib_struct calib;
static BMP_calib2_struct calib2;
static uint8_t ctrl_meas;		// Measurement setting

// --------------------------------------------------------------------------
uint8_t bmp_write(enum e_bmp_register bmp_reg, uint8_t data) {
	i2c_status_t tx_status;
	i2c_device_t i2c_devid = { BMP280_ADDR, BMP280_I2C_SPEED };	// Device address & 400 kbit/s

	tx_status = I2C_DRV_MasterSendDataBlocking(BMP280_I2C, &i2c_devid, &bmp_reg,
			sizeof(uint8_t), &data, sizeof(uint8_t), 30);

	return tx_status != kStatus_I2C_Success;
}

// --------------------------------------------------------------------------
uint8_t bmp_read(enum e_bmp_register bmp_reg, uint8_t *buffer) {
	i2c_status_t rx_status;
	i2c_device_t i2c_devid = { BMP280_ADDR, BMP280_I2C_SPEED };	// Device address & 400 kbit/s

	rx_status = I2C_DRV_MasterReceiveDataBlocking(BMP280_I2C, &i2c_devid,
			&bmp_reg, sizeof(uint8_t), buffer, sizeof(uint8_t), 30);

	return rx_status != kStatus_I2C_Success;
}

// --------------------------------------------------------------------------
uint8_t bmp_read_block(uint8_t bmp_reg_start, uint8_t *buffer, uint8_t length) {
	i2c_status_t rx_status;
	i2c_device_t i2c_devid = { BMP280_ADDR, BMP280_I2C_SPEED };	// Device address & 400 kbit/s

	rx_status = I2C_DRV_MasterReceiveDataBlocking(BMP280_I2C, &i2c_devid,
			&bmp_reg_start, sizeof(uint8_t), buffer, length, 30);

	return rx_status != kStatus_I2C_Success;
}

// --------------------------------------------------------------------------
uint8_t bmp_is_busy(void)
{
	uint8_t status;
	if (bmp_read(BMP_status, &status))		// I2C Reading error
		return -1;

	if (status & 0b00001001)		// Busy check
		return 1;					// Busy, converting in progress
	else
		return 0;
}

// --------------------------------------------------------------------------
uint8_t bmp_init(void) {
	uint8_t id;

	if (I2C_DRV_MasterInit(BMP280_I2C, &i2c) != kStatus_I2C_Success)
		return -1;

	bmp_read(BMP_id, &id);
	if (id != 0x60)
		return -1;			// ID not corresponding

	// osrs_t -> 0b001 = 1x temperature oversampling
	// osrs_p -> 0b011 = 4x pressure oversampling (standard resolution)
	// filter -> 0b000 = filter off
	// mode -> 0b01 = forced mode
	if (bmp_write(BMP_config, 0b00000000))
		return -1;

	ctrl_meas = 0b00101100;
	if (bmp_write(BMP_ctrl_meas, ctrl_meas))
		return -1;

	// Humidity overampling 1x
	if (bmp_write(BMP_ctrl_hum, 0b00000001))
		return -1;

	if (bmp_read_block((uint8_t)BMP_CAL_START_ADDR, (uint8_t *)&calib, sizeof(calib)))	// read calibration structure1
		return -1;

	uint8_t cal[7];
	if (bmp_read_block((uint8_t)BMP_CAL2_START_ADDR, cal, sizeof(cal)))	// read calibration structure2
		return -1;

	calib2.dig_H2 = cal[0] | (cal[1] << 8);
	calib2.dig_H3 = cal[2];
	calib2.dig_H4 = (cal[3] << 4) | (cal[4] & 0x0F);
	calib2.dig_H5 = (cal[5] << 4) | (cal[4] >> 4);
	calib2.dig_H6 = cal[6];
	return 0;
}

// --------------------------------------------------------------------------
// Returns temperature in DegC, resolution is 0.01 DegC. Output value of 5123 equals 51.23 DegC.
// t_fine carries fine temperature as global value
int32_t t_fine;
int32_t bmp280_compensate_T_int32(int32_t adc_T) {
	int32_t var1, var2, T;
	var1 = ((((adc_T >> 3) - ((int32_t)calib.dig_T1<<1))) * ((int32_t)calib.dig_T2)) >> 11;
	var2 = (((((adc_T >> 4) - ((int32_t)calib.dig_T1)) * ((adc_T>>4) - ((int32_t)calib.dig_T1))) >> 12) * ((int32_t)calib.dig_T3)) >> 14;
	t_fine = var1 + var2;
	T = (t_fine * 5 + 128) >> 8;
	return T;
}

// --------------------------------------------------------------------------
// Returns pressure in Pa as unsigned 32 bit integer in Q24.8 format (24 integer bits and 8 fractional bits).
// Output value of 24674867 represents 24674867/256 = 96386.2 Pa = 963.862 hPa
uint32_t bmp280_compensate_P_int64(int32_t adc_P) {
	int64_t var1, var2, p;
	var1 = ((int64_t) t_fine) - 128000;
	var2 = var1 * var1 * (int64_t) calib.dig_P6;
	var2 = var2 + ((var1 * (int64_t) calib.dig_P5) << 17);
	var2 = var2 + (((int64_t) calib.dig_P4) << 35);
	var1 = ((var1 * var1 * (int64_t) calib.dig_P3) >> 8) + ((var1 * (int64_t) calib.dig_P2) << 12);
	var1 = (((((int64_t) 1) << 47) + var1)) * ((int64_t) calib.dig_P1) >> 33;
	if (var1 == 0) {
		return 0; // avoid exception caused by division by zero
	}
	p = 1048576 - adc_P;
	p = (((p << 31) - var2) * 3125) / var1;
	var1 = (((int64_t) calib.dig_P9) * (p >> 13) * (p >> 13)) >> 25;
	var2 = (((int64_t) calib.dig_P8) * p) >> 19;
	p = ((p + var1 + var2) >> 8) + (((int64_t) calib.dig_P7) << 4);
	return (uint32_t) p;
}

// --------------------------------------------------------------------------
// Returns humidity in %RH as unsigned 32 bit integer in Q22.10 format (22 integer and 10 fractional bits).
// Output value of 47445 represents 47445/1024 = 46.333 %RH
uint32_t bme280_compensate_H_int32(int32_t adc_H)
{
	int32_t v_x1_u32r;
	v_x1_u32r = (t_fine - ((int32_t)76800));
	v_x1_u32r = (((((adc_H << 14) - (((int32_t)calib2.dig_H4) << 20) - (((int32_t)calib2.dig_H5) * v_x1_u32r)) +
	((int32_t)16384)) >> 15) * (((((((v_x1_u32r * ((int32_t)calib2.dig_H6)) >> 10) * (((v_x1_u32r *
	((int32_t)calib2.dig_H3)) >> 11) + ((int32_t)32768))) >> 10) + ((int32_t)2097152)) *
	((int32_t)calib2.dig_H2) + 8192) >> 14));
	v_x1_u32r = (v_x1_u32r - (((((v_x1_u32r >> 15) * (v_x1_u32r >> 15)) >> 7) * ((int32_t)calib.dig_H1)) >> 4));
	v_x1_u32r = (v_x1_u32r < 0 ? 0 : v_x1_u32r);
	v_x1_u32r = (v_x1_u32r > 419430400 ? 419430400 : v_x1_u32r);
	return (uint32_t)(v_x1_u32r>>12);
}


// --------------------------------------------------------------------------
uint8_t bmp_start_measure(void) {
	if (bmp_write(BMP_ctrl_meas, ctrl_meas | 0b01))		// Forced measurement mode
		return -1;
	return 0;
}

// --------------------------------------------------------------------------
uint8_t bmp_read_data(int16_t * temp, uint16_t * humid, uint32_t * press) {
	uint8_t press_raw[3];
	uint8_t temp_raw[3];
	uint8_t hum_raw[2];

	static int32_t temp_c;
	static uint32_t press_c;
	static uint32_t hum_c;

	if (bmp_is_busy())
		return 0;

	if (bmp_read_block(BMP_press_msb, press_raw, 3))
		return 0;

	if (bmp_read_block(BMP_temp_msb, temp_raw, 3))
		return 0;

	if (bmp_read_block(BMP_hum_msb, hum_raw, 2))
		return 0;

	temp_c = (int32_t)(((uint32_t)temp_raw[2] >> 4) | ((uint32_t)temp_raw[1] << 4) | ((uint32_t)temp_raw[0] << 12));
	press_c = ((uint32_t)press_raw[2] >> 4) | ((uint32_t)press_raw[1] << 4) | ((uint32_t)press_raw[0] << 12);
	hum_c = (uint32_t)hum_raw[1] | ((uint32_t)hum_raw[0] << 8);

	temp_c = bmp280_compensate_T_int32(temp_c);
	press_c = bmp280_compensate_P_int64(press_c);
	hum_c = bme280_compensate_H_int32(hum_c);

	*temp = (int16_t)temp_c;
	*press = press_c >> 8;
	*humid = (hum_c * 100) >> 10;

	return 1;
}



