#include "config.h"
#include "e_ink_display.h"
#include "pindef.h"
#include "main.h"

#include <spi.h>
#include <stddef.h>
#include <delay.h>

const unsigned char lut_full_update[]  =
{
	0x02, 0x02, 0x01, 0x11, 0x12, 0x12, 0x22, 0x22,
	0x66, 0x69, 0x69, 0x59, 0x58, 0x99, 0x99, 0x88,
	0x00, 0x00, 0x00, 0x00, 0xF8, 0xB4, 0x13, 0x51,
	0x35, 0x51, 0x51, 0x19, 0x01, 0x00
};

const unsigned char lut_partial_update[] =
{
	0x10, 0x18, 0x18, 0x08, 0x18, 0x18, 0x08, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x13, 0x14, 0x44, 0x12,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00
};

void Epd_SendCommand(unsigned char cmd)
{
	EINK_CMD();
	EINK_SELECT();
	delay_us(10);
	spi_write(cmd);
	EINK_UNSELECT();
}

void Epd_SendData(unsigned char data)
{
	EINK_DATA();
	EINK_SELECT();
	delay_us(10);
	spi_write(data);
	EINK_UNSELECT();
}

/**
 *  @brief: Wait until the busy_pin goes LOW
 */
void Epd_WaitUntilIdle(void) {
    while(EINK_BUSY) {      //LOW: idle, HIGH: busy
    	wait_to_interrupt();
    }      
}

/**
 *  @brief: module reset.
 *          often used to awaken the module in deep sleep,
 *          see Epd::Sleep();
 */
void Epd_Reset(void) {
    EINK_RST(0);   
    delay(10);
    EINK_RST(1); 
    delay(10);
}

/**
 *  @brief: set the look-up table register
 */
void Epd_SetLut(const unsigned char* lut) {
    Epd_SendCommand(ECMD_WRITE_LUT_REG);
    /* the length of look-up table is 30 bytes */
    for (int i = 0; i < 30; i++) {
        Epd_SendData(*(lut+i));
    }
}

/**
 *  @brief: put an image buffer to the frame memory.
 *          this won't update the display.
 */
void Epd_SetFrameMemoryPart(
    unsigned char* image_buffer,
    int x,
    int y,
    int image_width,
    int image_height
) {
    int x_end;
    int y_end;

    if (
        image_buffer == NULL ||
        x < 0 || image_width < 0 ||
        y < 0 || image_height < 0
    ) {
        return;
    }
    /* x point must be the multiple of 8 or the last 3 bits will be ignored */
    x &= 0xF8;
    image_width &= 0xF8;
    if (x + image_width >= EINK_WIDTH) {
        x_end = EINK_WIDTH - 1;
    } else {
        x_end = x + image_width - 1;
    }
    if (y + image_height >= EINK_HEIGHT) {
        y_end = EINK_HEIGHT - 1;
    } else {
        y_end = y + image_height - 1;
    }
    Epd_SetMemoryArea(x, y, x_end, y_end);
    Epd_SetMemoryPointer(x, y);
    Epd_SendCommand(ECMD_WRITE_RAM);
    /* send the image data */
    for (int j = 0; j < y_end - y + 1; j++) {
        for (int i = 0; i < (x_end - x + 1) / 8; i++) {
            Epd_SendData(image_buffer[i + j * (image_width / 8)]);
        }
    }
}

/**
 *  @brief: put an image buffer to the frame memory.
 *          this won't update the display.
 *
 *          Question: When do you use this function instead of 
 *          void SetFrameMemory(
 *              const unsigned char* image_buffer,
 *              int x,
 *              int y,
 *              int image_width,
 *              int image_height
 *          );
 *          Answer: SetFrameMemory with parameters only reads image data
 *          from the RAM but not from the flash in AVR chips (for AVR chips,
 *          you have to use the function pgm_read_byte to read buffers 
 *          from the flash).
 */
void Epd_SetFrameMemory(const unsigned char* image_buffer) {
    Epd_SetMemoryArea(0, 0, EINK_WIDTH - 1, EINK_HEIGHT - 1);
    Epd_SetMemoryPointer(0, 0);
    Epd_SendCommand(ECMD_WRITE_RAM);
    /* send the image data */
    for (int i = 0; i < EINK_WIDTH / 8 * EINK_HEIGHT; i++) {
        Epd_SendData(image_buffer[i]);
    }
}

/**
 *  @brief: clear the frame memory with the specified color.
 *          this won't update the display.
 */
void Epd_ClearFrameMemory(unsigned char color) {
    Epd_SetMemoryArea(0, 0, EINK_WIDTH - 1, EINK_HEIGHT - 1);
    Epd_SetMemoryPointer(0, 0);
    Epd_SendCommand(ECMD_WRITE_RAM);
    /* send the color data */
    for (int i = 0; i < EINK_WIDTH / 8 * EINK_HEIGHT; i++) {
        Epd_SendData(color);
    }
}

/**
 *  @brief: update the display
 *          there are 2 memory areas embedded in the e-paper display
 *          but once this function is called,
 *          the the next action of SetFrameMemory or ClearFrame will 
 *          set the other memory area.
 */
void Epd_DisplayFrame(void) {
    Epd_SendCommand(ECMD_DISPLAY_UPDATE_CTRL2);
    Epd_SendData(0xC4);
    Epd_SendCommand(ECMD_MASTER_ACTIVATION);
    Epd_SendCommand(ECMD_NOP);
    // Sleep until refresh is completed
    Epd_WaitUntilIdle();
}

/**
 *  @brief: private function to specify the memory area for data R/W
 */
void Epd_SetMemoryArea(int x_start, int y_start, int x_end, int y_end) {
    Epd_SendCommand(ECMD_SET_RAM_X_START_END);
    /* x point must be the multiple of 8 or the last 3 bits will be ignored */
    Epd_SendData((x_start >> 3) & 0xFF);
    Epd_SendData((x_end >> 3) & 0xFF);
    Epd_SendCommand(ECMD_SET_RAM_Y_START_END);
    Epd_SendData(y_start & 0xFF);
    Epd_SendData((y_start >> 8) & 0xFF);
    Epd_SendData(y_end & 0xFF);
    Epd_SendData((y_end >> 8) & 0xFF);
}

/**
 *  @brief: private function to specify the start point for data R/W
 */
void Epd_SetMemoryPointer(int x, int y) {
    Epd_SendCommand(ECMD_SET_RAM_X);
    /* x point must be the multiple of 8 or the last 3 bits will be ignored */
    Epd_SendData((x >> 3) & 0xFF);
    Epd_SendCommand(ECMD_SET_RAM_Y);
    Epd_SendData(y & 0xFF);
    Epd_SendData((y >> 8) & 0xFF);
    Epd_WaitUntilIdle();
}

/**
 *  @brief: After this command is transmitted, the chip would enter the 
 *          deep-sleep mode to save power. 
 *          The deep sleep mode would return to standby by hardware reset. 
 *          You can use Epd::Init() to awaken
 */
void Epd_Sleep() {
    Epd_SendCommand(ECMD_DEEP_SLEEP);
	Epd_SendData(0x01);
    // Epd_WaitUntilIdle();  Cannot be used because Pin goes high
}

/**
 *  @brief: Initialize Epaper display
 */
int Epd_Init(const unsigned char* lut) {
	
	SPI_Master_Init(1000000);
    /* EPD hardware init start */
    Epd_Reset();
    Epd_SendCommand(ECMD_DRIVE_OUT);
	Epd_SendData((EINK_HEIGHT - 1) & 0xFF);
    Epd_SendData(((EINK_HEIGHT - 1) >> 8) & 0xFF);
    Epd_SendData(0x00);                     // GD = 0; SM = 0; TB = 0;
    Epd_SendCommand(ECMD_BOOST_SOFT_START);
    Epd_SendData(0xD7);
    Epd_SendData(0xD6);
    Epd_SendData(0x9D);
    Epd_SendCommand(ECMD_WRITE_VCOM);
    Epd_SendData(0xA8);                     // VCOM 7C
    Epd_SendCommand(ECMD_SET_DUMMY_LINE_PER);
    Epd_SendData(0x1A);                     // 4 dummy lines per gate
    Epd_SendCommand(ECMD_SET_GATE_LINE_WIDTH);
    Epd_SendData(0x08);                     // 2us per line
    Epd_SendCommand(ECMD_DATA_ENTRY_MODE);
    Epd_SendData(0x03);                     // X increment; Y increment
    Epd_SetLut(lut);
    /* EPD hardware init end */
    return 0;
}
